# -*- coding: utf-8 -*-
# XXX: please be careful to only save this file with an utf-8 editor

from os import remove, rmdir, mkdir
from os.path import join, dirname, isdir
import unittest
from zipfile import ZipFile

import pytest

from kivy.clock import Clock
from kivy.uix.filechooser import FileChooserListView
from kivy.utils import platform

unicode_char = chr


class FileChooserUnicodeTestCase(unittest.TestCase):
    def setUp(self):
        basepath = dirname(__file__)
        subdir = join(basepath, "filechooser_files")
        self.subdir = subdir

        ufiles = [
            "कीवीtestu",
            "कीवीtestu" + unicode_char(0xEEEE),
            "कीवीtestu" + unicode_char(0xEEEE - 1),
            "कीवीtestu" + unicode_char(0xEE),
        ]
        self.files = [join(subdir, f) for f in ufiles]
        if not isdir(subdir):
            mkdir(subdir)
        for f in self.files:
            open(f, "wb").close()

        # existing files
        existfiles = [
            "à¤•à¥€à¤µà¥€test",
            "à¤•à¥€à¤’µà¥€test",
            "Ã Â¤â€¢Ã Â¥â‚¬Ã Â¤ÂµÃ Â¥â‚¬test",
            "testl\ufffe",
            "testl\uffff",
        ]
        self.existfiles = [join(subdir, f) for f in existfiles]
        with ZipFile(join(basepath, "unicode_files.zip"), "r") as myzip:
            myzip.extractall(path=subdir)
        for f in self.existfiles:
            open(f, "rb").close()

    @pytest.fixture(autouse=True)
    def set_clock(self, kivy_clock):
        self.kivy_clock = kivy_clock

    @pytest.mark.skipif(
        platform == "macosx" or platform == "ios",
        reason="Unicode files unpredictable on MacOS and iOS",
    )
    # On macOS and iOS, files ending in \uffff etc. are changed.
    # We cannot predict the filenames that will be created.
    # If it works on Window and Linux, we can safely assume it also works
    # on macOS and iOS.
    def test_filechooserlistview_unicode(self):

        wid = FileChooserListView(path=self.subdir)
        Clock.tick()
        files = [join(self.subdir, f) for f in wid.files]
        for f in self.files:
            self.assertIn(f, files)

        for f in self.existfiles:
            self.assertIn(f, files)

    def tearDown(self):
        for f in self.files + self.existfiles:
            try:
                remove(f)
            except (OSError, FileNotFoundError):
                pass
        try:
            rmdir(self.subdir)
        except (OSError, FileNotFoundError):
            pass



import json

import pytest

from kivy.config import ConfigParser
from kivy.uix.settings import Settings


def test_settings_create_json_panel_errors():
    config = ConfigParser()

    with pytest.raises(
        Exception, match="You must specify either the filename or data"
    ):
        Settings().create_json_panel("Demo", config, filename=None, data=None)

    with pytest.raises(
        ValueError, match="The first element must be a list"
    ):
        data = json.dumps({"key": "value"})
        Settings().create_json_panel("Demo", config, filename=None, data=data)

    with pytest.raises(
        ValueError, match="One setting are missing the \"type\" element"
    ):
        data = json.dumps([{"key": "value"}])
        Settings().create_json_panel("Demo", config, filename=None, data=data)

    with pytest.raises(
        ValueError, match="No class registered to handle the <testunknown> type"
    ):
        data = json.dumps([{"type": "testunknown"}])
        Settings().create_json_panel("Demo", config, filename=None, data=data)



from kivy.graphics.instructions cimport Instruction

cdef get_stencil_state()
cdef void restore_stencil_state(dict state)
cdef void reset_stencil_state()

cdef class StencilPush(Instruction):
    cdef bint _clear_stencil
    cdef bint _check_bool(self, object value)
    cdef int apply(self) except -1

cdef class StencilPop(Instruction):
    cdef int apply(self) except -1

cdef class StencilUse(Instruction):
    cdef int _op
    cdef int apply(self) except -1

cdef class StencilUnUse(Instruction):
    cdef int apply(self) except -1



from kivy.tests.common import GraphicUnitTest


class ClipboardTestCase(GraphicUnitTest):

    def setUp(self):
        from kivy.core.clipboard import Clipboard
        self._clippy = Clipboard
        clippy_types = Clipboard.get_types()
        cliptype = clippy_types[0]
        if 'UTF8_STRING' in clippy_types:
            cliptype = 'UTF8_STRING'
        self._cliptype = cliptype
        super(ClipboardTestCase, self).setUp()

    def test_clipboard_not_dummy(self):
        clippy = self._clippy
        if clippy.__class__.__name__ == 'ClipboardDummy':
            self.fail('Something went wrong "dummy" clipboard is being used')

    def test_clipboard_paste(self):
        clippy = self._clippy
        try:
            clippy.paste()
        except:
            self.fail(
                'Can not get data from clipboard')

    def test_clipboard_copy(self):
        clippy = self._clippy
        try:
            clippy.copy(u"Hello World")
        except:
            self.fail(
                'Can not put data to clipboard')

    def test_clipboard_copy_paste(self):
        clippy = self._clippy
        txt1 = u"Hello 1"
        clippy.copy(txt1)
        ret = clippy.paste()
        self.assertEqual(txt1, ret)

    def test_clipboard_copy_paste_with_emoji(self):
        clippy = self._clippy
        test_emoji_str = 'kivy 😀 😁 🤣 😃 😄 😅 😆 😉 😊 😋 😎 😍 😘 😗'
        clippy.copy(test_emoji_str)
        self.assertEqual(test_emoji_str, clippy.paste())



'''
Clipboard windows: an implementation of the Clipboard using ctypes.
'''

__all__ = ('ClipboardWindows', )

from kivy.utils import platform
from kivy.core.clipboard import ClipboardBase

if platform != 'win':
    raise SystemError('unsupported platform for Windows clipboard')

import ctypes
from ctypes import wintypes
user32 = ctypes.windll.user32
kernel32 = ctypes.windll.kernel32
msvcrt = ctypes.cdll.msvcrt
c_char_p = ctypes.c_char_p
c_wchar_p = ctypes.c_wchar_p

GlobalLock = kernel32.GlobalLock
GlobalLock.argtypes = [wintypes.HGLOBAL]
GlobalLock.restype = wintypes.LPVOID

GlobalUnlock = kernel32.GlobalUnlock
GlobalUnlock.argtypes = [wintypes.HGLOBAL]
GlobalUnlock.restype = wintypes.BOOL

CF_UNICODETEXT = 13
GMEM_MOVEABLE = 0x0002


class ClipboardWindows(ClipboardBase):

    def _copy(self, data):
        self._ensure_clipboard()
        self.put(data, self._clip_mime_type)

    def get(self, mimetype='text/plain'):
        GetClipboardData = user32.GetClipboardData
        GetClipboardData.argtypes = [wintypes.UINT]
        GetClipboardData.restype = wintypes.HANDLE

        user32.OpenClipboard(user32.GetActiveWindow())

        # GetClipboardData returns a HANDLE to the clipboard data
        # which is a memory object containing the data
        # See: https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getclipboarddata  # noqa: E501
        pcontents = GetClipboardData(CF_UNICODETEXT)

        # if someone pastes a FILE, the content is None for SCF 13
        # and the clipboard is locked if not closed properly
        if not pcontents:
            user32.CloseClipboard()
            return ''

        # The handle returned by GetClipboardData is a memory object
        # and needs to be locked to get the actual pointer to the data
        pcontents_locked = GlobalLock(pcontents)
        data = c_wchar_p(pcontents_locked).value
        GlobalUnlock(pcontents)

        user32.CloseClipboard()
        return data

    def put(self, text, mimetype="text/plain"):
        SetClipboardData = user32.SetClipboardData
        SetClipboardData.argtypes = [wintypes.UINT, wintypes.HANDLE]
        SetClipboardData.restype = wintypes.HANDLE

        GlobalAlloc = kernel32.GlobalAlloc
        GlobalAlloc.argtypes = [wintypes.UINT, ctypes.c_size_t]
        GlobalAlloc.restype = wintypes.HGLOBAL

        user32.OpenClipboard(user32.GetActiveWindow())
        user32.EmptyClipboard()

        # The wsclen function returns the number of
        # wide characters in a string (not including the null character)
        text_len = msvcrt.wcslen(text) + 1

        # According to the docs regarding SetClipboardDatam, if the hMem
        # parameter identifies a memory object, the object must have
        # been allocated using the GMEM_MOVEABLE flag.
        # See: https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setclipboarddata  # noqa: E501
        # The size of the memory object is the number of wide characters in
        # the string plus one for the terminating null character
        hCd = GlobalAlloc(
            GMEM_MOVEABLE, ctypes.sizeof(ctypes.c_wchar) * text_len
        )

        # Since the memory object is allocated with GMEM_MOVEABLE, should be
        # locked to get the actual pointer to the data.
        hCd_locked = GlobalLock(hCd)
        ctypes.memmove(
            c_wchar_p(hCd_locked),
            c_wchar_p(text),
            ctypes.sizeof(ctypes.c_wchar) * text_len,
        )
        GlobalUnlock(hCd)

        # Finally, set the clipboard data (and then close the clipboard)
        SetClipboardData(CF_UNICODETEXT, hCd)
        user32.CloseClipboard()

    def get_types(self):
        return ['text/plain']
